@@ -6,6 +6,7 @@
 #include <NTL/ZZ.h>
 #include <NTL/WordVector.h>
 #include <NTL/vec_GF2.h>
+#include <NTL/Lazy.h>
 
 NTL_OPEN_NNS
 
@@ -31,9 +32,37 @@ explicit GF2X(GF2 a) { *this = a; }
 
 ~GF2X() { }
 
+// we have to explicitly declare copy constructor and assignment
+// in case we also declare the corresponding move operations
+
+GF2X(const GF2X& a) : xrep(a.xrep) { }
+GF2X& operator=(const GF2X& a) { xrep = a.xrep; return *this; }
+
+#if (NTL_CXX_STANDARD >= 2011)
+
+GF2X(GF2X&& a) NTL_FAKE_NOEXCEPT
+{
+   *this = std::move(a);
+}
+
+GF2X& operator=(GF2X&& a) NTL_FAKE_NOEXCEPT
+{
+   if (xrep.pinned() || a.xrep.pinned()) {
+      *this = a;
+   }
+   else {
+      xrep.unpinned_move(a.xrep);
+   }
+
+   return *this;
+}
+
+
+#endif
+
+
 GF2X(INIT_SIZE_TYPE, long n);
 
-GF2X& operator=(const GF2X& a) { xrep = a.xrep; return *this; }
 
 inline GF2X& operator=(GF2 a);
 inline GF2X& operator=(long a);
@@ -49,13 +78,37 @@ void SetMaxLength(long n);
 
 
 void SetLength(long n);
-ref_GF2 operator[](long i);
-const GF2 operator[](long i) const;
+
+ref_GF2 operator[](long i) 
+{
+#ifdef NTL_RANGE_CHECK
+   if (i < 0) LogicError("index out of range in Vec");
+   long q, r;
+   _ntl_bpl_divrem(cast_unsigned(i), q, r);
+   if (q >= xrep.length()) LogicError("index out of range in Vec");
+#endif
+
+   vec_GF2::iterator t(INIT_LOOP_HOLE, xrep.elts(), i);
+   return *t;
+}
+
+
+const GF2 operator[](long i) const
+{
+#ifdef NTL_RANGE_CHECK
+   if (i < 0) LogicError("index out of range in Vec");
+   long q, r;
+   _ntl_bpl_divrem(cast_unsigned(i), q, r);
+   if (q >= xrep.length()) LogicError("index out of range in Vec");
+#endif
+
+   vec_GF2::const_iterator t(INIT_LOOP_HOLE, xrep.elts(), i);
+   return *t;
+}
 
 
 
-
-NTL_THREAD_LOCAL static long HexOutput;
+static NTL_CHEAP_THREAD_LOCAL long HexOutput;
 
 inline GF2X(long i, GF2 c);
 inline GF2X(long i, long c);
@@ -69,14 +122,22 @@ GF2X(GF2X& x, INIT_TRANS_TYPE) : xrep(x.xrep, INIT_TRANS) { }
 // that are not be subject to special memory management.
 
 
+void swap(GF2X& x) { xrep.swap(x.xrep); }
+
+
+
+
 
 // mainly for internal consumption by GF2XWatcher
 
-void release() { xrep.release(); }
+void KillBig() { xrep.KillBig(); }
 
 };
 
 
+NTL_DECLARE_RELOCATABLE((GF2X*))
+
+
 long IsZero(const GF2X& a);
 
 long IsOne(const GF2X& a); 
@@ -110,7 +171,7 @@ inline GF2X::GF2X(INIT_MONO_TYPE, long i, long a) { SetCoeff(*this, i, a); }
 inline GF2X::GF2X(INIT_MONO_TYPE, long i) { SetCoeff(*this, i); }
 
 
-void swap(GF2X& a, GF2X& b);
+inline void swap(GF2X& a, GF2X& b) { a.swap(b); }
 
 long deg(const GF2X& aa);
 
@@ -335,7 +396,6 @@ class GF2XModulus {
 
 public:
    GF2XModulus();
-   ~GF2XModulus();
 
    GF2XModulus(const GF2XModulus&);  
    GF2XModulus& operator=(const GF2XModulus&); 
@@ -364,17 +424,19 @@ public:
    long method; 
 
    vec_GF2X stab;
-   _ntl_ulong **stab_ptr;
-   long *stab_cnt;
 
-   _ntl_ulong *stab1;
+   UniqueArray<unsigned long *> stab_ptr;
+   UniqueArray<long> stab_cnt;
+   UniqueArray<unsigned long> stab1;
+
 
    GF2X h0, f0;
 
-   vec_GF2 tracevec;
+   OptionalVal< Lazy<vec_GF2> > tracevec;
 
 }; 
 
+NTL_DECLARE_RELOCATABLE((GF2XModulus*))
 
 inline long deg(const GF2XModulus& F) { return F.n; }
 
@@ -725,17 +787,31 @@ inline long NumBytes(const GF2X& a)
 
 class GF2XWatcher {
 public:
-   GF2X *watched;
+   GF2X& watched;
    explicit
-   GF2XWatcher(GF2X *_watched) : watched(_watched) {}
+   GF2XWatcher(GF2X& _watched) : watched(_watched) {}
 
-   ~GF2XWatcher() { watched->release(); } 
+   ~GF2XWatcher() { watched.KillBig(); } 
 };
 
-#define NTL_GF2XRegister(x) NTL_THREAD_LOCAL static GF2X x; GF2XWatcher _WATCHER__ ## x(&x)
+#define NTL_GF2XRegister(x) NTL_TLS_LOCAL(GF2X, x); GF2XWatcher _WATCHER__ ## x(x)
 
 
 
+// RAII for HexOutput
+
+class GF2XHexOutputPush {
+private:
+   long OldHexOutput;
+
+   GF2XHexOutputPush(const GF2XHexOutputPush&); // disable
+   void operator=(const GF2XHexOutputPush&); // disable
+
+public:
+   GF2XHexOutputPush() : OldHexOutput(GF2X::HexOutput) { }
+   ~GF2XHexOutputPush() { GF2X::HexOutput = OldHexOutput; }
+};
+
 
 NTL_CLOSE_NNS
 
