@@ -29,7 +29,7 @@ NTL_OPEN_NNS
 #ifndef NTL_RANGE_CHECK
 #define NTL_WV_RANGE_CHECK_CODE 
 #else
-#define NTL_WV_RANGE_CHECK_CODE if (i < 0 || !rep || i >= long(rep[-1])) RangeError(i);
+#define NTL_WV_RANGE_CHECK_CODE if (i < 0 || !rep || i >= long(rep[-1])) LogicError("index out of range in WordVector");
 #endif
 
 // vectors are allocated in chunks of this size
@@ -41,7 +41,7 @@ NTL_OPEN_NNS
 // vectors are always expanded by at least this ratio
 
 #ifndef NTL_WordVectorExpansionRatio
-#define NTL_WordVectorExpansionRatio (1.2)
+#define NTL_WordVectorExpansionRatio (1.4)
 #endif
 
 // controls initialization during input
@@ -54,7 +54,6 @@ NTL_OPEN_NNS
 class WordVector {  
 public:  
    _ntl_ulong *rep;  
-   void RangeError(long i) const;  
 
    WordVector(WordVector& x, INIT_TRANS_TYPE) { rep = x.rep; x.rep = 0; }
 
@@ -66,10 +65,29 @@ public:
 
    WordVector& operator=(const WordVector& a);  
 
+   bool pinned() const 
+   {
+      return rep && (rep[-2] & 1);
+   }
+
+   // assumes *this and other are not pinned
+   void unpinned_swap(WordVector& other)
+   {
+      _ntl_swap(this->rep, other.rep);
+   }
+
+   // assumes *this and other are not pinned
+   void unpinned_move(WordVector& other)
+   {
+      WordVector tmp;
+      tmp.unpinned_swap(other);
+      tmp.unpinned_swap(*this);
+   }
+
    ~WordVector();  
    void kill(); 
 
-   void release() { if (MaxLength() > NTL_RELEASE_THRESH) kill(); }
+   void KillBig() { if (MaxLength() > NTL_RELEASE_THRESH) kill(); }
    // this conditinally kills the vector, if its size is excessive
 
    void DoSetLength(long n);
@@ -111,19 +129,32 @@ public:
    const _ntl_ulong* elts() const { return rep; }  
    _ntl_ulong* elts() { return rep; }  
          
-   static void swap_impl(WordVector& x, WordVector& y);  
-   static void append_impl(WordVector& v, _ntl_ulong a); 
-   static void append_impl(WordVector& v, const WordVector& w); 
+   void swap(WordVector& y);  
+   void append(_ntl_ulong a); 
+   void append(const WordVector& w); 
 }; 
 
+
+
+
+class WordVectorWatcher {
+public:
+   WordVector& watched;
+   explicit
+   WordVectorWatcher(WordVector& _watched) : watched(_watched) {}
+
+   ~WordVectorWatcher() { watched.KillBig(); }
+};
+
+
 inline void swap(WordVector& x, WordVector& y) 
-   { WordVector::swap_impl(x, y); }
+   { x.swap(y); }
 
 inline void append(WordVector& v, _ntl_ulong a)
-   { WordVector::append_impl(v, a); }
+   { v.append(a); }
 
 inline void append(WordVector& v, const WordVector& w)
-   { WordVector::append_impl(v, w); }
+   { v.append(w); }
 
 
 NTL_SNS istream& operator>>(NTL_SNS istream&, WordVector&);  
