@@ -34,14 +34,8 @@ ZZX(INIT_SIZE_TYPE, long n)
 
    { rep.SetMaxLength(n); }
 
-ZZX(const ZZX& a) : rep(a.rep) { }
-// initial value is a
-
-
-ZZX& operator=(const ZZX& a) 
-   { rep = a.rep; return *this; }
-
-~ZZX() { }
+// default copy constructor and assignment
+// default destructor
 
 void normalize();
 // strip leading zeros
@@ -84,9 +78,15 @@ inline ZZX& operator=(const ZZ& a);
 
 ZZX(ZZX& x, INIT_TRANS_TYPE) : rep(x.rep, INIT_TRANS) { }
 
+void swap(ZZX& x) { rep.swap(x.rep); }
+// swap with x (only pointers are swapped)
+
 };
 
 
+NTL_DECLARE_RELOCATABLE((ZZX*))
+
+
 
 
 /********************************************************************
@@ -171,7 +171,7 @@ inline void set(ZZX& x)
 inline void swap(ZZX& x, ZZX& y)
 // swap x & y (only pointers are swapped)
 
-   { swap(x.rep, y.rep); }
+   { x.swap(y); }
 
 void trunc(ZZX& x, const ZZX& a, long m);
 // x = a % X^m
