@@ -12,60 +12,56 @@ NTL_OPEN_NNS
 
 class ZZ_pEInfoT {
 private:
+
    ZZ_pEInfoT();                       // disabled
    ZZ_pEInfoT(const ZZ_pEInfoT&);   // disabled
    void operator=(const ZZ_pEInfoT&);  // disabled
 public:
-   long ref_count;
 
    ZZ_pEInfoT(const ZZ_pX&);
    ~ZZ_pEInfoT() { }
 
    ZZ_pXModulus p;
 
-   ZZ   _card;
-   long _card_init;
    ZZ   _card_base;
    long _card_exp;
 
+   Lazy<ZZ>  _card;
+
 
 };
 
-NTL_THREAD_LOCAL extern ZZ_pEInfoT *ZZ_pEInfo; // info for current modulus, initially null
+extern
+NTL_CHEAP_THREAD_LOCAL 
+ZZ_pEInfoT *ZZ_pEInfo; 
+// info for current modulus, initially null
+// raw pointer for faster TLS access
 
 
-// FIXME: in a thread-safe impl, we will need to use shared_ptr's 
-// for reference counted pointers
 
 
 
 class ZZ_pEContext {
 private:
-ZZ_pEInfoT *ptr;
+SmartPtr<ZZ_pEInfoT> ptr;
 
 public:
+
+ZZ_pEContext() { }
+explicit ZZ_pEContext(const ZZ_pX& p) : ptr(MakeSmart<ZZ_pEInfoT>(p)) { }
+
+// copy constructor, assignment, destructor: default
+
 void save();
 void restore() const;
 
-ZZ_pEContext() { ptr = 0; }
-explicit ZZ_pEContext(const ZZ_pX& p);
-
-ZZ_pEContext(const ZZ_pEContext&); 
-
-
-ZZ_pEContext& operator=(const ZZ_pEContext&); 
-
-
-~ZZ_pEContext();
-
-
 };
 
 
 class ZZ_pEBak {
 private:
-long MustRestore;
-ZZ_pEInfoT *ptr;
+ZZ_pEContext c;
+bool MustRestore;
 
 ZZ_pEBak(const ZZ_pEBak&); // disabled
 void operator=(const ZZ_pEBak&); // disabled
@@ -74,7 +70,7 @@ public:
 void save();
 void restore();
 
-ZZ_pEBak() { MustRestore = 0; ptr = 0; }
+ZZ_pEBak() : MustRestore(false) {  }
 
 ~ZZ_pEBak();
 
@@ -83,6 +79,8 @@ ZZ_pEBak() { MustRestore = 0; ptr = 0; }
 
 
 
+
+
 class ZZ_pEPush {
 private:
 ZZ_pEBak bak;
@@ -131,21 +129,19 @@ explicit ZZ_pE(long a) { *this = a;  } // NO_ALLOC
 explicit ZZ_pE(const ZZ_p& a) { *this = a;  } // NO_ALLOC
 
 
-ZZ_pE(const ZZ_pE& a)  {  _ZZ_pE__rep = a._ZZ_pE__rep; } // NO_ALLOC
 
 ZZ_pE(INIT_NO_ALLOC_TYPE) { }  // allocates no space
 ZZ_pE(INIT_ALLOC_TYPE) {_ZZ_pE__rep.rep.SetMaxLength(ZZ_pE::degree());  }  // allocates space
 void allocate() { _ZZ_pE__rep.rep.SetMaxLength(ZZ_pE::degree()); }
 
-~ZZ_pE() { } 
-
-ZZ_pE& operator=(const ZZ_pE& a) { _ZZ_pE__rep = a._ZZ_pE__rep; return *this; }
 
 inline ZZ_pE& operator=(long a);
 inline ZZ_pE& operator=(const ZZ_p& a);
 
 ZZ_pE(ZZ_pE& x, INIT_TRANS_TYPE) : _ZZ_pE__rep(x._ZZ_pE__rep, INIT_TRANS) { }
 
+void swap(ZZ_pE& x) { _ZZ_pE__rep.swap(x._ZZ_pE__rep); }
+
 
 // You can always access the _ZZ_pE__representation directly...if you dare.
 ZZ_pX& LoopHole() { return _ZZ_pE__rep; }
@@ -163,10 +159,15 @@ static long initialized() { return (ZZ_pEInfo != 0); }
 static void init(const ZZ_pX&);
 
 
+
 };
 
 
 
+NTL_DECLARE_RELOCATABLE((ZZ_pE*))
+
+
+
 
 // read-only access to _ZZ_pE__representation
 inline const ZZ_pX& rep(const ZZ_pE& a) { return a._ZZ_pE__rep; }
@@ -182,7 +183,7 @@ inline void set(ZZ_pE& x)
 inline void swap(ZZ_pE& x, ZZ_pE& y)
 // swap x and y
 
-   { swap(x._ZZ_pE__rep, y._ZZ_pE__rep); }
+   { x.swap(y); }
 
 // ****** addition
 
