@@ -8,6 +8,13 @@
 #include <NTL/vec_ZZ.h>
 #include <NTL/vec_ZZ_p.h>
 #include <NTL/FFT.h>
+#include <NTL/Lazy.h>
+#include <NTL/SmartPtr.h>
+
+#ifndef NTL_WIZARD_HACK
+#include <NTL/mat_ZZ_p.h>
+#endif
+
 
 NTL_OPEN_NNS
 
@@ -79,14 +86,9 @@ explicit ZZ_pX(const ZZ_p& a) { *this = a; }
 
 ZZ_pX(INIT_SIZE_TYPE, long n) { rep.SetMaxLength(n); }
 
-ZZ_pX(const ZZ_pX& a) : rep(a.rep) { }
-// initial value is a
+// default copy constrctor and assignment
 
-
-ZZ_pX& operator=(const ZZ_pX& a) 
-   { rep = a.rep; return *this; }
-
-~ZZ_pX() { }
+// default destructor
 
 void normalize();
 // strip leading zeros
@@ -125,10 +127,20 @@ ZZ_pX& operator=(long a);
 ZZ_pX& operator=(const ZZ_p& a);
 
 
+void swap(ZZ_pX& x)
+{
+   rep.swap(x.rep);
+}
+
+
 };
 
 
 
+NTL_DECLARE_RELOCATABLE((ZZ_pX*))
+
+
+
 
 /********************************************************************
 
@@ -211,7 +223,7 @@ inline void set(ZZ_pX& x)
 inline void swap(ZZ_pX& x, ZZ_pX& y)
 // swap x & y (only pointers are swapped)
 
-   { swap(x.rep, y.rep); }
+   { x.swap(y); }
 
 void random(ZZ_pX& x, long n);
 inline ZZ_pX random_ZZ_pX(long n)
@@ -550,18 +562,20 @@ class FFTRep {
 public:
    long k;                // a 2^k point representation
    long MaxK;             // maximum space allocated
-   long **tbl;
    long NumPrimes; 
+   Unique2DArray<long> tbl;
 
-   void SetSize(long NewK);
+   FFTRep() : k(-1), MaxK(-1), NumPrimes(0) { }
+
+   FFTRep(const FFTRep& R) : k(-1), MaxK(-1), NumPrimes(0)
+   { *this = R; }
+
+   FFTRep(INIT_SIZE_TYPE, long InitK) : k(-1), MaxK(-1), NumPrimes(0)
+   {  SetSize(InitK); }
 
-   FFTRep(const FFTRep& R);
    FFTRep& operator=(const FFTRep& R);
-
-   FFTRep() { k = MaxK = -1; tbl = 0; NumPrimes = 0; }
-   FFTRep(INIT_SIZE_TYPE, long InitK) 
-   { k = MaxK = -1; tbl = 0; NumPrimes = 0; SetSize(InitK); }
-   ~FFTRep();
+   void SetSize(long NewK);
+   void DoSetSize(long NewK, long NewNumPrimes);
 };
 
 
@@ -629,15 +643,14 @@ private:
 public:
    long n;
    long MaxN;
-   long **tbl;
    long NumPrimes; 
+   Unique2DArray<long> tbl;
 
    void SetSize(long NewN);
 
-   ZZ_pXModRep() { n = MaxN = 0; tbl = 0; NumPrimes = 0; }
-   ZZ_pXModRep(INIT_SIZE_TYPE, long k) 
-   { n = MaxN = 0; tbl = 0; NumPrimes = 0; SetSize(k); }
-   ~ZZ_pXModRep();
+   ZZ_pXModRep() : n(0), MaxN(0), NumPrimes(0) {  }
+   ZZ_pXModRep(INIT_SIZE_TYPE, long NewN) : n(0), MaxN(0), NumPrimes(0)
+   {  SetSize(NewN); }
 };
 
 
@@ -648,6 +661,12 @@ void ToFFTRep(FFTRep& x, const ZZ_pXModRep& a, long k, long lo, long hi);
 // must have hi-lo+1 < 2^k
 
 
+void FromFFTRep(ZZ_pXModRep& x, const FFTRep& a);
+// for testing and timing purposes only -- converts from FFTRep
+
+void FromZZ_pXModRep(ZZ_pX& x, const ZZ_pXModRep& a, long lo, long hi);
+// for testing and timing purposes only -- converts from ZZ_pXModRep
+
 
 
 
@@ -864,7 +883,10 @@ public:
    FFTRep FRep; // 2^k point rep of f
                 // H = rev((rev(f))^{-1} rem X^{n-1})
    FFTRep HRep; // 2^l point rep of H
-   vec_ZZ_p tracevec;  // mutable
+
+   OptionalVal< Lazy<vec_ZZ_p> > tracevec;
+   // an extra level of indirection to ensure the class
+   // can be used in a Vec (there may be a mutex in the Lazy object)
 
    // but these will remain public
    ZZ_pXModulus(const ZZ_pX& ff);
@@ -874,6 +896,9 @@ public:
 
 };
 
+
+NTL_DECLARE_RELOCATABLE((ZZ_pXModulus*))
+
 inline long deg(const ZZ_pXModulus& F) { return F.n; }
 
 void build(ZZ_pXModulus& F, const ZZ_pX& f);
@@ -1090,7 +1115,7 @@ struct ZZ_pXArgument {
    vec_ZZ_pX H;
 };
 
-NTL_THREAD_LOCAL extern long ZZ_pXArgBound;
+extern NTL_CHEAP_THREAD_LOCAL long ZZ_pXArgBound;
 
 
 void build(ZZ_pXArgument& H, const ZZ_pX& h, const ZZ_pXModulus& F, long m);
@@ -1105,6 +1130,40 @@ CompMod(const ZZ_pX& g, const ZZ_pXArgument& H, const ZZ_pXModulus& F)
    { ZZ_pX x; CompMod(x, g, H, F); NTL_OPT_RETURN(ZZ_pX, x); }
 
 
+// New alternative CompMod strategy that just reduces to 
+// matrix multiplication ... 
+
+#if (!defined(NTL_WIZARD_HACK) && defined(NTL_HAVE_LL_TYPE))
+// NOTE: currently, we only ise the multi-modular matmul NTL_HAVE_LL_TYPE
+// if NTL_HAVE_LL_TYPE is defined.
+
+#define NTL_USE_ZZ_pXNewArgument
+
+#endif
+
+#ifdef NTL_USE_ZZ_pXNewArgument
+
+struct ZZ_pXNewArgument {
+   mat_ZZ_p_opaque mat;
+   ZZ_pX poly;
+};
+
+void build(ZZ_pXNewArgument& H, const ZZ_pX& h, const ZZ_pXModulus& F, long m);
+void CompMod(ZZ_pX& x, const ZZ_pX& g, const ZZ_pXNewArgument& H,
+             const ZZ_pXModulus& F);
+
+void ProjectPowers(vec_ZZ_p& x, const vec_ZZ_p& a, long k,
+                   const ZZ_pXNewArgument& H, const ZZ_pXModulus& F);
+
+#else
+
+typedef ZZ_pXArgument ZZ_pXNewArgument;
+
+#endif
+
+
+
+
 #ifndef NTL_TRANSITION
 
 void UpdateMap(vec_ZZ_p& x, const vec_ZZ_p& a,
