@@ -4,14 +4,42 @@
 
 #include <NTL/config.h>
 #include <NTL/mach_desc.h>
-#include <NTL/HAVE_LL_TYPE.h>
-#include <NTL/HAVE_BUILTIN_CLZL.h>
-#include <NTL/HAVE_AVX.h>
-#include <NTL/HAVE_FMA.h>
+
+#include <NTL/ALL_FEATURES.h>
+
+
+// defines the working C++ standard
+
+#if defined(NTL_STD_CXX11)
+#define NTL_CXX_STANDARD (2011)
+#elif defined(NTL_STD_CXX14)
+#define NTL_CXX_STANDARD (2014)
+#else
+#define NTL_CXX_STANDARD (1998)
+#endif
+
+// define some macros regarding noexcept declarations
+
+#if (NTL_CXX_STANDARD >= 2011)
+
+#define NTL_NOEXCEPT noexcept
+
+#ifdef NTL_EXCEPTIONS
+#define NTL_FAKE_NOEXCEPT
+#else
+#define NTL_FAKE_NOEXCEPT noexcept
+#endif
+
+#else
+
+#define NTL_NOEXCEPT 
+#define NTL_FAKE_NOEXCEPT
+
+#endif
 
 
 /*
- * Resolve double-word integer types.
+ * Resolve double-word integer type.
  *
  * Unfortunately, there is no "standard" way to do this.
  * On 32-bit machines, 'long long' usually works (but not
@@ -20,37 +48,12 @@
  * non-standard double-word type.  
  *
  * Note that C99 creates a standard header <stdint.h>,
- * but it is not clear how widely this is implemented yet,
- * and for example, GCC does not provide a type int128_t 
+ * but it is not clear how widely this is implemented,
+ * and for example, older versions of GCC does not provide a type int128_t 
  * in <stdint.h> on 64-bit machines.
  */
 
 
-#if (defined(NTL_LONG_LONG_TYPE))
-
-#define NTL_LL_TYPE NTL_LONG_LONG_TYPE
-
-#elif (NTL_BITS_PER_LONG == 64 && defined(__GNUC__))
-
-#define NTL_LL_TYPE __int128_t
-
-#elif (NTL_BITS_PER_LONG == 32 && (defined(_MSC_VER) || defined(__BORLANDC__)))
-
-#define NTL_LL_TYPE __int64
-
-#elif (NTL_BITS_PER_LONG == 64 && (defined(_MSC_VER) || defined(__BORLANDC__)))
-
-#define NTL_LL_TYPE __int128
-
-#endif
-
-#if (!defined(NTL_LL_TYPE))
-
-#define NTL_LL_TYPE long long
-
-#endif
-
-
 
 #if (defined(NTL_UNSIGNED_LONG_LONG_TYPE))
 
@@ -58,7 +61,7 @@
 
 #elif (NTL_BITS_PER_LONG == 64 && defined(__GNUC__))
 
-#define NTL_ULL_TYPE __uint128_t
+#define NTL_ULL_TYPE __uint128_t 
 
 #elif (NTL_BITS_PER_LONG == 32 && (defined(_MSC_VER) || defined(__BORLANDC__)))
 
@@ -79,17 +82,14 @@
 
 #ifdef NTL_HAVE_LL_TYPE
 
-typedef NTL_LL_TYPE _ntl_longlong;
 typedef NTL_ULL_TYPE _ntl_ulonglong;
 // typenames are more convenient than macros
 
 #else
 
-#undef NTL_LL_TYPE
 #undef NTL_ULL_TYPE
 // prevent any use of these macros
 
-class _ntl_longlong { private: _ntl_longlong() { } };
 class _ntl_ulonglong { private: _ntl_ulonglong() { } };
 // cannot create variables of these types
 
@@ -111,7 +111,7 @@ typedef unsigned int _ntl_uint32; // 32-bit word
 
 #else
 
-// NOTE: C++ standard guarntees longs ar at least 32-bits wide,
+// NOTE: C++ standard guarantees longs are at least 32-bits wide,
 // and this is also explicitly checked at builod time
 
 typedef unsigned long _ntl_uint32; // 32-bit word
@@ -394,6 +394,83 @@ void _ntl_swap(T*& a, T*& b)
  */
 
 
+// The following do for "move" what the above does for swap
+
+#define NTL_DEFINE_SCALAR_MOVE(T)\
+inline T _ntl_scalar_move(T& a)\
+{\
+   T t = a; a = 0; return t;\
+}
+
+NTL_DEFINE_SCALAR_MOVE(long)
+NTL_DEFINE_SCALAR_MOVE(int)
+NTL_DEFINE_SCALAR_MOVE(short)
+NTL_DEFINE_SCALAR_MOVE(char)
+
+NTL_DEFINE_SCALAR_MOVE(unsigned long)
+NTL_DEFINE_SCALAR_MOVE(unsigned int)
+NTL_DEFINE_SCALAR_MOVE(unsigned short)
+NTL_DEFINE_SCALAR_MOVE(unsigned char)
+
+NTL_DEFINE_SCALAR_MOVE(double)
+NTL_DEFINE_SCALAR_MOVE(float)
+
+   
+template<class T>
+T* _ntl_scalar_move(T*& a)
+{
+   T *t = a; a = 0; return t;
+}
+
+
+
+
+
+// The following routine increments a pointer so that
+// it is properly aligned.  
+// It is assumed that align > 0.
+// If align is a constant power of 2, it compiles
+// into a small handful of simple instructions.
+
+#if (NTL_BIG_POINTERS)
+
+#define NTL_UPTRINT_T unsigned long long
+// DIRT: this should really be std::uintptr_t, defined
+// in <cstdint>; however, that header is not widely available,
+// and even if it were, std::uintptr_t is not guaranteed
+// to be defined.  Of course, unsigned long long may not
+// be defined in pre-C++11.  
+
+#else
+
+#define NTL_UPTRINT_T unsigned long 
+
+#endif
+
+
+#ifdef NTL_HAVE_ALIGNED_ARRAY
+
+inline
+char *_ntl_make_aligned(char *p, long align)
+{
+   unsigned long r =  (unsigned long) (((NTL_UPTRINT_T) (p)) % ((NTL_UPTRINT_T) (align)));
+   return p + ((((unsigned long) (align)) - r) % ((unsigned long) (align)));
+}
+
+#else
+
+
+inline
+char *_ntl_make_aligned(char *p, long align)
+{
+   return p;
+}
+
+
+#endif
+
+
+
 
 
 // The following is for aligning small local arrays
@@ -408,9 +485,7 @@ void _ntl_swap(T*& a, T*& b)
 
 #define NTL_ALIGNED_LOCAL_ARRAY(align, x, type, n) \
    char x##__ntl_hidden_variable_storage[n*sizeof(type)+align]; \
-   type *x = (type *) ((&x##__ntl_hidden_variable_storage[0]) + \
-                       ((-((unsigned long) (&x##__ntl_hidden_variable_storage[0]))) %\
-                       (unsigned long)(align))) \
+   type *x = (type *) _ntl_make_aligned(&x##__ntl_hidden_variable_storage[0], align);
 
 
 #define NTL_AVX_BYTE_ALIGN (32)
@@ -423,7 +498,101 @@ void _ntl_swap(T*& a, T*& b)
 // and it should also be as big as a cache line
 
 
-   
+
+#ifdef NTL_HAVE_BUILTIN_CLZL
+
+inline long 
+_ntl_count_bits(unsigned long x)
+{
+   return x ? (NTL_BITS_PER_LONG - __builtin_clzl(x)) : 0;
+}
+
+#else
+
+inline long 
+_ntl_count_bits(unsigned long x)
+{
+   if (!x) return 0;
+
+   long res = NTL_BITS_PER_LONG;
+   while (x < (1UL << (NTL_BITS_PER_LONG-1))) {
+      x <<= 1;
+      res--;
+   }
+
+   return res;
+}
 
 #endif
 
+
+
+
+#if (!defined(NTL_CLEAN_INT) && NTL_ARITH_RIGHT_SHIFT && (NTL_BITS_PER_LONG == (1 << (NTL_NUMBITS_BPL-1))))
+
+
+
+inline void
+_ntl_bpl_divrem(long a, long& q, long& r)
+{
+   q = a >> (NTL_NUMBITS_BPL-1);
+   r = a & (NTL_BITS_PER_LONG-1);
+}
+
+#else
+
+inline void
+_ntl_bpl_divrem(long a, long& q, long& r)
+{
+   q = a / NTL_BITS_PER_LONG;
+   r = a % NTL_BITS_PER_LONG;
+   if (r < 0) {
+      q--;
+      r += NTL_BITS_PER_LONG;
+   }
+}
+
+#endif
+
+inline void
+_ntl_bpl_divrem(unsigned long a, long& q, long& r)
+{
+   q = a / NTL_BITS_PER_LONG;
+   r = a % NTL_BITS_PER_LONG;
+}
+
+
+
+template <class T>
+struct _ntl_is_char_pointer
+{
+ enum {value = false};
+};
+
+template <>
+struct _ntl_is_char_pointer<char*>
+{
+ enum {value = true};
+};
+
+template <>
+struct _ntl_is_char_pointer<const char*>
+{
+ enum {value = true};
+};
+
+template <bool, typename T = void>
+struct _ntl_enable_if
+{};
+
+template <typename T>
+struct _ntl_enable_if<true, T> {
+  typedef T type;
+};
+
+
+
+
+
+
+#endif
