@@ -5,6 +5,9 @@
 #include <NTL/vector.h>
 #include <NTL/lzz_p.h>
 #include <NTL/vec_lzz_p.h>
+#include <NTL/Lazy.h>
+#include <NTL/SmartPtr.h>
+#include <NTL/mat_lzz_p.h>
 
 NTL_OPEN_NNS
 
@@ -84,8 +87,6 @@ explicit zz_pX(zz_p a) { *this = a; }
 
 zz_pX(INIT_SIZE_TYPE, long n) { rep.SetMaxLength(n); }
 
-zz_pX(const zz_pX& a) : rep(a.rep) { }
-// initial value is a
 
 inline zz_pX(long i, zz_p c);
 inline zz_pX(long i, long c);
@@ -94,13 +95,12 @@ inline zz_pX(INIT_MONO_TYPE, long i, zz_p c);
 inline zz_pX(INIT_MONO_TYPE, long i, long c);
 inline zz_pX(INIT_MONO_TYPE, long i);
 
-zz_pX& operator=(const zz_pX& a) 
-   { rep = a.rep; return *this; }
 
 inline zz_pX& operator=(long a);
 inline zz_pX& operator=(zz_p a);
 
-~zz_pX() { }
+// default copy constructor and assignment
+// default destructor
 
 void normalize();
 // strip leading zeros
@@ -124,6 +124,10 @@ zz_p& operator[](long i) { return rep[i]; }
 const zz_p& operator[](long i) const { return rep[i]; }
 
 
+void swap(zz_pX& x)
+{
+   rep.swap(x.rep);
+}
 
 
 static const zz_pX& zero();
@@ -133,6 +137,9 @@ zz_pX(zz_pX& x, INIT_TRANS_TYPE) : rep(x.rep, INIT_TRANS) { }
 };
 
 
+NTL_DECLARE_RELOCATABLE((zz_pX*))
+
+
 
 
 /********************************************************************
@@ -217,7 +224,7 @@ inline void set(zz_pX& x)
 inline void swap(zz_pX& x, zz_pX& y)
 // swap x & y (only pointers are swapped)
 
-   { swap(x.rep, y.rep); }
+   { x.swap(y); }
 
 void random(zz_pX& x, long n);
 inline zz_pX random_zz_pX(long n)
@@ -565,18 +572,20 @@ class fftRep {
 public:
    long k;                // a 2^k point representation
    long MaxK;             // maximum space allocated
-   long *tbl[4];
    long NumPrimes;
+   UniqueArray<long> tbl[4];
+
+   fftRep() : k(-1), MaxK(-1), NumPrimes(0) { }
+
+   fftRep(const fftRep& R) : k(-1), MaxK(-1), NumPrimes(0)
+   { *this = R; }
+
+   fftRep(INIT_SIZE_TYPE, long InitK) : k(-1), MaxK(-1), NumPrimes(0)
+   { SetSize(InitK); }
 
-   fftRep(const fftRep&); 
    fftRep& operator=(const fftRep&); 
-
    void SetSize(long NewK);
-
-   fftRep() { k = MaxK = -1; NumPrimes = 0; }
-   fftRep(INIT_SIZE_TYPE, long InitK) 
-   { k = MaxK = -1; NumPrimes = 0; SetSize(InitK); }
-   ~fftRep();
+   void DoSetSize(long NewK, long NewNumPrimes);
 };
 
 
@@ -846,7 +855,9 @@ public:
    fftRep FRep; // 2^k point rep of f
                 // H = rev((rev(f))^{-1} rem X^{n-1})
    fftRep HRep; // 2^l point rep of H
-   vec_zz_p tracevec;  // mutable
+
+   OptionalVal< Lazy<vec_zz_p> > tracevec;  
+   // extra level of indirection to ensure relocatability
 
    zz_pXModulus(const zz_pX& ff);
 
@@ -856,6 +867,9 @@ public:
 };
 
 
+NTL_DECLARE_RELOCATABLE((zz_pXModulus*))
+
+
 inline long deg(const zz_pXModulus& F) { return F.n; }
 
 void build(zz_pXModulus& F, const zz_pX& f);
@@ -1084,7 +1098,9 @@ struct zz_pXArgument {
    vec_zz_pX H;
 };
 
-NTL_THREAD_LOCAL extern long zz_pXArgBound;
+extern 
+NTL_CHEAP_THREAD_LOCAL 
+long zz_pXArgBound;
 
 
 void build(zz_pXArgument& H, const zz_pX& h, const zz_pXModulus& F, long m);
@@ -1099,6 +1115,28 @@ CompMod(const zz_pX& g, const zz_pXArgument& H, const zz_pXModulus& F)
    { zz_pX x; CompMod(x, g, H, F); NTL_OPT_RETURN(zz_pX, x); }
 
 
+// New alternative CompMod strategy that just reduces to 
+// matrix multiplication ... 
+
+
+struct zz_pXNewArgument {
+   Mat<zz_p> mat;
+   zz_pX     poly;
+};
+
+void build(zz_pXNewArgument& H, const zz_pX& h, const zz_pXModulus& F, long m);
+void CompMod(zz_pX& x, const zz_pX& g, const zz_pXNewArgument& H,
+             const zz_pXModulus& F);
+void reduce(zz_pXNewArgument& H, const zz_pXModulus& F);
+
+void ProjectPowers(vec_zz_p& x, const vec_zz_p& a, long k,
+                   const zz_pXNewArgument& H, const zz_pXModulus& F);
+
+
+
+
+
+
 
 #ifndef NTL_TRANSITION
 
